@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013,2018 MariaDB Corporation AB
+   Copyright (C) 2013,2019 MariaDB Corporation AB
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -29,10 +29,55 @@ void MADB_StmtResetResultStructures(MADB_Stmt *Stmt)
     sizeof(long) * mysql_stmt_field_count(Stmt->stmt));
   memset(Stmt->Lengths, 0, sizeof(long) * mysql_stmt_field_count(Stmt->stmt));
 
+  Stmt->LastRowFetched= 0;
   MADB_STMT_RESET_CURSOR(Stmt);
 }
 /* }}} */
 
+/* {{{ MoveNext - moves C/C cursor forward for Offset positions */
+SQLRETURN MoveNext(MADB_Stmt *Stmt, unsigned long long Offset)
+{
+  SQLRETURN  result= SQL_SUCCESS;
+
+  if (Stmt->result != NULL)
+  {
+    unsigned int i;
+    char        *SavedFlag;
+
+    SavedFlag= (char*)MADB_CALLOC(mysql_stmt_field_count(Stmt->stmt));
+
+    if (SavedFlag == NULL)
+    {
+      return SQL_ERROR;
+    }
+
+    for (i=0; i < mysql_stmt_field_count(Stmt->stmt); i++)
+    {
+      SavedFlag[i]= Stmt->stmt->bind[i].flags & MADB_BIND_DUMMY;
+
+      Stmt->stmt->bind[i].flags|= MADB_BIND_DUMMY;
+    }
+
+    while (Offset--)
+    {
+      if (mysql_stmt_fetch(Stmt->stmt) == 1)
+      {
+        result= SQL_ERROR;
+        break;
+      }
+    }
+
+    for (i=0; i < mysql_stmt_field_count(Stmt->stmt); i++)
+    {
+      Stmt->stmt->bind[i].flags &= (~MADB_BIND_DUMMY | SavedFlag[i]);
+    }
+
+    MADB_FREE(SavedFlag);
+  }
+  return result;
+}
+/* }}} */
+
 /* {{{ MADB_StmtDataSeek */
 SQLRETURN MADB_StmtDataSeek(MADB_Stmt *Stmt, my_ulonglong FetchOffset)
 {
@@ -49,15 +94,41 @@ SQLRETURN MADB_StmtDataSeek(MADB_Stmt *Stmt, my_ulonglong FetchOffset)
 }
 /* }}} */
 
+/* {{{  */
+void QuickDropAllPendingResults(MYSQL* Mariadb)
+{
+  int Next= 0;
+  do {
+    if (Next == 0)
+    {
+      if (mysql_field_count(Mariadb) > 0)
+      {
+        MYSQL_RES *Res= mysql_store_result(Mariadb);
+
+        if (Res)
+        {
+          mysql_free_result(Res);
+        }
+      }
+    }
+  } while ((Next= mysql_next_result(Mariadb)) != -1);
+}
+/* }}} */
+
 /* {{{ MADB_StmtMoreResults */
 SQLRETURN MADB_StmtMoreResults(MADB_Stmt *Stmt)
 {
   SQLRETURN ret= SQL_SUCCESS;
+
   if (!Stmt->stmt)
   {
     return MADB_SetError(&Stmt->Error, MADB_ERR_08S01, NULL, 0);
   }
 
+  /* We can't have it in MADB_StmtResetResultStructures, as it breaks dyn_cursor functionality.
+     Thus we free-ing bind structs on move to new result only */
+  MADB_FREE(Stmt->result);
+
   if (Stmt->MultiStmts)
   {
     if (Stmt->MultiStmtNr == STMT_COUNT(Stmt->Query) - 1)
@@ -67,7 +138,7 @@ SQLRETURN MADB_StmtMoreResults(MADB_Stmt *Stmt)
 
     ++Stmt->MultiStmtNr;
 
-    MADB_InstallStmt(Stmt);
+    MADB_InstallStmt(Stmt, Stmt->MultiStmts[Stmt->MultiStmtNr]);
 
     return SQL_SUCCESS;
   }
@@ -79,51 +150,79 @@ SQLRETURN MADB_StmtMoreResults(MADB_Stmt *Stmt)
       return SQL_NO_DATA;
     else
     {
+      int Next;
+
       LOCK_MARIADB(Stmt->Connection);
-      mysql_next_result(Stmt->Connection->mariadb);
+      Next= mysql_next_result(Stmt->Connection->mariadb);
+
+      if (Next > 0)
+      {
+        ret= MADB_SetError(&Stmt->Error, MADB_ERR_HY000, mysql_error(Stmt->Connection->mariadb), 0);
+      }
+      else if (mysql_field_count(Stmt->Connection->mariadb) != 0)
+      {
+        MYSQL_RES *Res= mysql_store_result(Stmt->Connection->mariadb);
+        if (Res != NULL)
+        {
+          mysql_free_result(Res);
+        }
+        ret= MADB_SetError(&Stmt->Error, MADB_ERR_01000, "Internal error - unexpected text result received", 0);
+      }
+      else
+      {
+        Stmt->AffectedRows= mysql_affected_rows(Stmt->Connection->mariadb);
+      }
       UNLOCK_MARIADB(Stmt->Connection);
     }
     return ret;
   }
 
   if (mysql_stmt_more_results(Stmt->stmt))
+  {
     mysql_stmt_free_result(Stmt->stmt);
+  }
   else
+  {
     return SQL_NO_DATA;
+  }
   
   LOCK_MARIADB(Stmt->Connection);
-  if (mysql_stmt_next_result(Stmt->stmt) ||
-      !Stmt->stmt->field_count)
+  if (mysql_stmt_next_result(Stmt->stmt) > 0)
   {
     UNLOCK_MARIADB(Stmt->Connection);
-    return SQL_NO_DATA;
+    return MADB_SetNativeError(&Stmt->Error, SQL_HANDLE_STMT, Stmt->stmt);
   }
-
+  
   MADB_StmtResetResultStructures(Stmt);
-  /* We can't have it in MADB_StmtResetResultStructures, as it breaks dyn_cursor functionality.
-     Thus we free-ing bind structs on move to new result only */
-  MADB_FREE(Stmt->result);
 
-  MADB_DescSetIrdMetadata(Stmt, mysql_fetch_fields(FetchMetadata(Stmt)), mysql_stmt_field_count(Stmt->stmt));
-
-  if (Stmt->Connection->mariadb->server_status & SERVER_PS_OUT_PARAMS)
+  if (mysql_stmt_field_count(Stmt->stmt) == 0)
   {
-    Stmt->State= MADB_SS_OUTPARAMSFETCHED;
-    ret= Stmt->Methods->GetOutParams(Stmt, 0);
+    MADB_DescFree(Stmt->Ird, TRUE);
+    Stmt->AffectedRows= mysql_stmt_affected_rows(Stmt->stmt);
   }
   else
   {
-    if (Stmt->Options.CursorType != SQL_CURSOR_FORWARD_ONLY)
-    {
-      mysql_stmt_store_result(Stmt->stmt);
-    }
-  }
-  UNLOCK_MARIADB(Stmt->Connection);
+    unsigned int ServerStatus;
 
-  if (/* mysql_stmt_field_count(Stmt->stmt) && */
-        Stmt->Options.CursorType != SQL_CURSOR_FORWARD_ONLY)
-  {
-    mysql_stmt_data_seek(Stmt->stmt, 0);
+    MADB_DescSetIrdMetadata(Stmt, mysql_fetch_fields(FetchMetadata(Stmt)), mysql_stmt_field_count(Stmt->stmt));
+    Stmt->AffectedRows= 0;
+
+    mariadb_get_infov(Stmt->Connection->mariadb, MARIADB_CONNECTION_SERVER_STATUS, (void*)&ServerStatus);
+
+    if (ServerStatus & SERVER_PS_OUT_PARAMS)
+    {
+      Stmt->State= MADB_SS_OUTPARAMSFETCHED;
+      ret= Stmt->Methods->GetOutParams(Stmt, 0);
+    }
+    else
+    {
+      if (Stmt->Options.CursorType != SQL_CURSOR_FORWARD_ONLY)
+      {
+        mysql_stmt_store_result(Stmt->stmt);
+        mysql_stmt_data_seek(Stmt->stmt, 0);
+      }
+    }
+    UNLOCK_MARIADB(Stmt->Connection);
   }
 
   return ret;
@@ -139,9 +238,18 @@ SQLULEN MADB_RowsToFetch(MADB_Cursor *Cursor, SQLULEN ArraySize, unsigned long l
   Cursor->RowsetSize= ArraySize;
 
   if (Position + ArraySize > RowsInResultst)
-    result= (SQLULEN)(RowsInResultst - Position);
+  {
+    if (Position >= 0 && RowsInResultst > Position)
+    {
+      result= (SQLULEN)(RowsInResultst - Position);
+    }
+    else
+    {
+      result= 1;
+    }
+  }
 
-  return result >= 0 ? result : 1;
+  return result;
 }
 /* }}} */
 
